/*  Copyright (C) 2000, 2006  Marc Toussaint (mtoussai@inf.ed.ac.uk)
    under the terms of the GNU LGPL (http://www.gnu.org/copyleft/lesser.html)
    see the `std.h' file for a full copyright statement  */

/*! \file array.h
  \brief core file: defines the MT::Array class */

#ifndef MT_array_h
#define MT_array_h

#ifdef SWIG
%{
#include "array.h"
  %}
#endif

#include"std.h"

#define FOR1D(x,i)   for(i=0;i<x.d0;i++)
#define FOR2D(x,i,j) for(i=0;i<x.d0;i++) for(j=0;j<x.d1;j++)
#define forAll(i,A)  for(i=A.p;i!=A.pstop;i++)

typedef unsigned char byte;
typedef unsigned int uint;


//===========================================================================
//
// Array class
//

namespace MT{
  /*!\brief Simple array container to store arbitrary-dimensional arrays
    (tensors); can buffer more memory than necessary for faster
    resize; enables non-const reference of subarrays; enables fast
    memove for elementary types; implements many standard
    array/matrix/tensor operations. Please see the fully public attributes at the
    bottom of this page -- everthing is meant to be perfectly
    transparent. Interfacing with ordinary C-buffers is simple,
    e.g. via \c Array::referTo (Cbuffer,size) and \c Array::p and \c
    Array::pp. Please see also the reference for the \ref array.h
    header, which contains lots of functions that can be applied on
    Arrays. */

  template<class T>
  class Array{
  public:
    typedef Array<T> ArrayT;

    T *p;     //!< the pointer on the linear memory allocated
    uint N;   //!< number of elements
    uint nd;  //!< number of dimensions
    uint d0;  //!< 0th dim
    uint d1;  //!< 1st dim
    uint d2;  //!< 2nd dim
    uint *d;  //!< pointer to dimensions (for nd<3 points to d0)
    T *pstop; //!< end of memory (pstop is already out of the bound)
    uint M;   //!< size of allocated memory (may be greater than N)
    bool reference;//!< true if this refers to some external memory
    T **pp;   //!< C-style 2D pointer (only valid if created with \c Array::getCarray!)
    Array<uint> *sparse; //!< to allocate element, column and row index lists (1+d0+d1 lists)
    bool temp;

    enum matrixType{ full, diag };
    matrixType mtype;
    static int sizeT;

    //options
    /*!\brief be careful!!!; this option will modify the \c
      Array::resizeCopy : instead of calling the copy operator= for
      each element, the memory is moved with the \c memmove command -
      works only if you know what you do ... [default is true for
      simple types (like double, int, etc) and false for unknown
      ones] */
    bool memMove;

    /*!\brief if flexiMem is true (which is default!) the resize method will
      (1) at the first call allocate the exact amount of memory, (2)
      at further calls of increasing memory allocate twice the memory
      needed or (3) at further calls of decreasing memory only free
      the memory if the new size is smaller than a fourth */
    bool flexiMem;
    //@}

  private:
    static char memMoveInit;
    void init();

  public:
    //constructors
    Array();
    Array(const ArrayT& a);
    Array(uint i);
    Array(uint i,uint j);
    Array(uint i,uint j,uint k);
    Array(uint i,uint j,uint k,uint l);
    Array(const ArrayT& a,uint dim);
    Array(const ArrayT& a,uint i,uint j);
    Array(T* p,uint size);

    ~Array();
    ArrayT& operator=(const T v);
    ArrayT& operator=(const ArrayT& a);

    //resizing
    ArrayT& resize    (uint D0);
    ArrayT& resizeCopy(uint D0);
    ArrayT& reshape   (uint D0);
    ArrayT& resize    (uint D0,uint D1);
    ArrayT& resizeCopy(uint D0,uint D1);
    ArrayT& reshape   (uint D0,uint D1);
    ArrayT& resize    (uint D0,uint D1,uint D2);
    ArrayT& resizeCopy(uint D0,uint D1,uint D2);
    ArrayT& reshape   (uint D0,uint D1,uint D2);
    ArrayT& resize    (uint D0,uint D1,uint D2,uint D3);
    void resize    (uint ND,uint *dim);
    void resizeCopy(uint ND,uint *dim);
    void reshape   (uint ND,uint *dim);
    void resizeAs    (const ArrayT& a);
    void resizeCopyAs(const ArrayT& a);
    void reshapeAs   (const ArrayT& a);

    //initializing/setting entries
    void clear();
    void setText(const char* str);
    void setZero(byte zero=0);
    void setUni(const T& x,int d=-1);
    void setId(int d=-1);
    void setDiag(const T& x,int d=-1);
    void setBlock(const ArrayT& A,const ArrayT& B,const ArrayT& C,const ArrayT& D);
    void setBlock(const ArrayT& B,uint lo0,uint lo1);
    void setBlock(const ArrayT& B,uint lo);
    void setStraightPerm(int n=-1);
    void setCarray(T *buffer,uint D0);
    void setCarray(T **buffer,uint D0,uint D1);
    void referTo(const T *buffer,uint n);
    void referTo(const ArrayT& a);
    void referToSubRange(const ArrayT& a,uint i,int I);
    void referToSubDim(const ArrayT& a,uint dim);
    void referToSubDim(const ArrayT& a,uint i,uint j);
    void redirect(const ArrayT& a,uint i);
    void takeOver(ArrayT& a);

    //access
    T& elem(uint i) const;
    T& scalar() const;
    T& last() const;
    T& operator()(uint i) const;
    T& operator()(uint i,uint j) const;
    T& operator()(uint i,uint j,uint k) const;
    ArrayT operator[](uint i) const;
    ArrayT& operator()();
    ArrayT sub(uint i,int I) const;
    ArrayT sub(uint i,int I,uint j,int J) const;
    ArrayT sub(uint i,int I,uint j,int J,uint k,int K) const;
    ArrayT sub(uint i,int I,Array<uint> cols) const;
    T** getCarray() const;
    void getBlock(ArrayT& B,uint lo0,uint lo1) const;
    void getBlock(ArrayT& B,uint lo) const;
    T& max() const;
    T& min() const;
    uint maxIndex() const;
    void maxIndex(uint& i,uint& j) const;
    void maxIndex(uint& i,uint& j,uint& k) const;
    uint minIndex() const;
    int contains(const T& x) const;
    uint getMemsize() const;
    void copyInto  (T *buffer) const;
    void copyInto2D(T **buffer) const;

    //appending etc
    T& append();
    void append(const T& x);
    void append(const ArrayT& x);
    void append(const T *p,uint n);
    // appends only if x is not in array yet
    void setAppend(const T& x);
    void setAppend(const ArrayT& x);
    void replicate(uint copies);
    void insert(uint i,const T& x);
    bool setInsert(const T& x);
    void remove(uint i,uint n=1);
    void removePerm(uint i);
    uint findValue(const T& x) const;
    void removeValue(const T& x);
    bool removeValueSafe(const T& x);
    void replace(uint i,uint n,const ArrayT& x);
    void delRow(uint i);
    void delColumns(uint i,uint k=1);
    void insColumns(uint i,uint k=1);
    void insRows(uint i,uint k=1);
    T popLast(){ T x; x=elem(N-1); remove(N-1); return x; }

    //sorting etc
    typedef bool (*ElemCompare)(const T& a,const T& b);
    uint findInSorted(const T& x,ElemCompare comp){
      if(!N) return 0;
      if(comp(x,elem(0))) return 0;
      if(!comp(x,elem(N-1))) return N;
      uint lo=0,up=N-1,mi;
      for(;;){
        if(lo+1==up) return up;
        mi=lo+(up-lo)/2;
	if(comp(x,elem(mi))) up=mi; else lo=mi;
      }
      HALT("you shouldn't be here!");
      return 0;
    }
    uint insertSorted(const T& x,ElemCompare comp){
      CHECK(memMove,"");
      if(!N){ append(x); return 1; }
      uint i=findInSorted(x,comp);
      insert(i,x);
      return i;
    }
    void removeValueInSorted(const T& x,ElemCompare comp){
      uint i=findInSorted(x,comp);
      CHECK(elem(i)==x,"value not found");
      remove(i);
    }

    //the array as permutation
    void permute(uint i,uint j);
    void randomPermutation();
    void initPermutation(uint n);
    void randomPermutation(uint n);
    void reversePermutation(uint n);
    void push(int offset,bool circle=true);

    //tensor operations [to be moved out of class declaration]
    void tensorResize(const Array<uint> &newD);
    void tensorReshape(const Array<uint> &newD);
    void tensorAddRange(uint slot);
    void tensorCondNormalize(int left);
    void tensorCheckCondNormalization(uint left,double tol=1e-10);
    void tensorCondMax(uint left);
    void tensorCondSoftMax(uint left,double beta);
    T& operator()(const Array<uint> &I) const;
    void getIndexTuple(Array<uint> &I,uint i) const;
    uint D(uint k) const;

    //sparse matrices [to be moved outside]
    double sparsity();
    void makeSparse();

    //I/O
    void write(std::ostream& os,const char *ELEMSEP=" ",const char *LINESEP="\n ",bool binary=false) const;
    void read(std::istream& is,bool binary=false);
    void readOld(std::istream& is);
    void writeDim(std::ostream& os) const;
    void readDim(std::istream& is);
    const ArrayT& ioraw() const;
    const ArrayT& ionoraw() const;
    void readRaw(std::istream& is,uint d1=0,uint d2=0);
    void readRaw(const char* filename,uint d1,uint d2);
    void writeRaw(const char* filename);

    //kind of private
    void resizeMEM(uint n,bool copy);
    void freeMEM();
    void resetD();
  };
}


//===========================================================================
//
//!@name standard types
//

typedef MT::Array<double> arr;
typedef MT::Array<double> doubleA;
typedef MT::Array<float>  floatA;
typedef MT::Array<uint>   uintA;
typedef MT::Array<int>    intA;
typedef MT::Array<char>   charA;
typedef MT::Array<byte>   byteA;
typedef MT::Array<bool>   boolA;


//===========================================================================
//
// non-template functions
//

MT::Array<double> Identity(uint dim);
void makeSymmetric(MT::Array<double>& A);
void transpose(MT::Array<double>& A);
void SUS(const arr& p,uint n,uintA& s);
uint SUS(const arr& p);

namespace MT{
  //! use this to turn on Lapack routines [default true if MT_LAPACK is defined]
  extern bool useLapack;
}


//matrix-matrix multiplication
void blasMM(void *pX,void *pA,void *pB);

uint lapackSVD(const arr& A, arr& U, arr& d, arr& Vt);
void lapackEigenDecomp(const arr& symmA, arr& Evals, arr& Evecs);
uint own_svd_implementation(const arr& A, arr& U, arr& w, arr& V, bool sort);

uint svd(const arr& A,arr& U,arr& d,arr& V,bool sort=true);
void svd(const arr& A,arr& U,arr& V);
uint inverse_SVD(arr& inverse,const arr& A);

namespace MT{  void inverse_LU(arr& invX,const arr& X); }

uint inverse(arr& inverse,const arr& A);
void pseudoInverse(arr& inverse,const arr& A,const arr& W,double robustnessEps=0.);
arr inverse(const arr& A);
double determinant(const arr& A);
double cofactor(const arr& A,uint i,uint j);
bool isPositiveSemiDefinite(const arr& symmA);

void getIndexTuple(uintA &I,uint i,const uintA &d);
void tensor(arr &X, const arr &A, const uintA &pickA, const arr &B, const uintA &pickB, uint sum=0);
void lognormScale(arr& P,double& logP,bool force=true);


void gnuplot(const arr& X);
void write(const arr& X,const arr& Y,const char* name);
void write(const arr& X,const arr& Y,const arr& Z,const char* name);
void write(const arr& X,const arr& Y,const arr& Z,const arr& W,const char* name);

void write_ppm(const byteA &img,const char *file_name,bool swap_rows=false);
void read_ppm(byteA &img,const char *file_name,bool swap_rows=false);
void add_alpha_channel(byteA &img,byte alpha);
void make_grey(byteA &img);
void make_RGB(byteA &img);

void scanArrFile(const char* name);


//===========================================================================
//
// template functions
//

template<class T,class S> void resizeAs    (MT::Array<T>& x,const MT::Array<S>& a);
template<class T,class S> void resizeCopyAs(MT::Array<T>& x,const MT::Array<S>& a);
template<class T,class S> void reshapeAs   (MT::Array<T>& x,const MT::Array<S>& a);
template<class T,class S> void copy        (MT::Array<T>& x,const MT::Array<S>& a);
/*{
  resizeAs(a);
  uint i;
  if(memMove && typeid(T)==typeid(S)) memmove(p,a.p,sizeT*N);
  else for(i=0;i<N;i++) p[i]=(T)a.p[i];
  return *this;
  }*/

//! permutes the entries of a according to the current this
template<class T> void permute(MT::Array<T>& a,const MT::Array<uint>& permutation){
  CHECK(permutation.N<=a.N,"array smaller than permutation ("<<a.N<<"<"<<permutation.N<<")");
  MT::Array<T> b=a;
  for(uint i=0;i<a.N;i++) a.elem(i)=b.elem(permutation(i));
}
//! permutes the entries of a according to the inverse of the current this
template<class T> void invpermute(MT::Array<T>& a,const MT::Array<uint>& permutation){
  CHECK(permutation.N<=a.N,"array smaller than permutation ("<<a.N<<"<"<<permutation.N<<")");
  MT::Array<T> b=a;
  for(uint i=0;i<a.N;i++) a.elem(permutation(i))=b.elem(i);
}
//! randomly permute all entries of a
template<class T> void randomPermute(MT::Array<T>& a){
  MT::Array<uint> perm;
  perm.randomPermutation(a.N);
  permute(a,perm);
}

template<class T> void transpose(MT::Array<T>& x,const MT::Array<T>& y);
template<class T> void diag(MT::Array<T>& x,const MT::Array<T>& y);
template<class T> void setDiagonal(MT::Array<T>& x,const MT::Array<T>& v);
template<class T> void inverse2d(MT::Array<T>& invA,const MT::Array<T>& A);
template<class T> void blockMatrix(MT::Array<T>& X,const MT::Array<T>& A,const MT::Array<T>& B,const MT::Array<T>& C,const MT::Array<T>& D);
template<class T> void blockVector(MT::Array<T>& x,const MT::Array<T>& a,const MT::Array<T>& b);
template<class T> void blockMatrix(MT::Array<T>& X,const MT::Array<T>& A,const MT::Array<T>& B,const MT::Array<T>& C,const T& D);

template<class T> T entropy(const MT::Array<T>& v);
template<class T> T normalizeDist(MT::Array<T>& v);
template<class T> void makeConditional(MT::Array<T>& P);
template<class T> void checkNormalization(MT::Array<T>& v,double tol);
template<class T> void checkNormalization(MT::Array<T>& v){ checkNormalization(v,1e-10); }
template<class T> void eliminate(MT::Array<T>& x,const MT::Array<T>& y,uint d);
template<class T> void eliminate(MT::Array<T>& x,const MT::Array<T>& y,uint d,uint e);
template<class T> void eliminatePartial(MT::Array<T>& x,const MT::Array<T>& y,uint d);

template<class T> T sqrDistance(const MT::Array<T>& v, const MT::Array<T>& w);
template<class T> T maxDiff(const MT::Array<T>& v,const MT::Array<T>& w,uint *maxi=0);
template<class T> T maxRelDiff(const MT::Array<T>& v,const MT::Array<T>& w,T tol);
//template<class T> T sqrDistance(const MT::Array<T>& v, const MT::Array<T>& w,const MT::Array<bool>& mask);
template<class T> T sqrDistance(const MT::Array<T>& g,const MT::Array<T>& v,const MT::Array<T>& w);
template<class T> T euclideanDistance(const MT::Array<T>& v, const MT::Array<T>& w);
template<class T> T metricDistance(const MT::Array<T>& g,const MT::Array<T>& v, const MT::Array<T>& w);

template<class T> T sum(const MT::Array<T>& v);
template<class T> T sumOfAbs(const MT::Array<T>& v);
template<class T> T sumOfSqr(const MT::Array<T>& v);
template<class T> T norm(const MT::Array<T>& v);
template<class T> T mean(const MT::Array<T>& v);

template<class T> T var(const MT::Array<T>& v);
template<class T> T trace(const MT::Array<T>& v);

template<class T> T minDiag(const MT::Array<T>& v);

template<class T> T product(const MT::Array<T>& v);
template<class T> void innerProduct(MT::Array<T>& x,const MT::Array<T>& y, const MT::Array<T>& z);
template<class T> void outerProduct(MT::Array<T>& x,const MT::Array<T>& y, const MT::Array<T>& z);
template<class T> T scalarProduct(const MT::Array<T>& v, const MT::Array<T>& w);
template<class T> T scalarProduct(const MT::Array<T>& g,const MT::Array<T>& v, const MT::Array<T>& w);

template<class T> void tensorMarginal    (MT::Array<T> &Y,const MT::Array<T> &X,const uintA &Yid);
template<class T> void tensorMaxMarginal (MT::Array<T> &Y,const MT::Array<T> &X,const uintA &Yid);
template<class T> void tensorMarginal_old(MT::Array<T> &y,const MT::Array<T> &x,const uintA &xd,const uintA &ids);
template<class T> void tensorMultiply    (MT::Array<T> &X,const MT::Array<T> &Y,const uintA &Yid);
template<class T> void tensorAdd         (MT::Array<T> &X,const MT::Array<T> &Y,const uintA &Yid);
template<class T> void tensorMultiply_old(MT::Array<T> &x,const MT::Array<T> &y,const uintA &d,const uintA &ids);
template<class T> void tensorDivide      (MT::Array<T> &X,const MT::Array<T> &Y,const uintA &Yid);

template<class T> void setUnite(MT::Array<T>& x,const MT::Array<T>& y, const MT::Array<T>& z);
template<class T> void setSection(MT::Array<T>& x,const MT::Array<T>& y, const MT::Array<T>& z);
template<class T> void setMinus(MT::Array<T>& x,const MT::Array<T>& y);
template<class T> uint numberSharedElements(const MT::Array<T>& x,const MT::Array<T>& y);
template<class T> T minA(const MT::Array<T>& v);
template<class T> T minA(const MT::Array<T>& v, uint & ind, uint start=0, uint end=0);
template<class T> T maxA(const MT::Array<T>& v);
template<class T> T maxA(const MT::Array<T>& v, uint & ind, uint start=0, uint end=0);
template<class T> void minmaxA(const MT::Array<T>& v, T& minVal, T& maxVal);
template<class T> T absMax(const MT::Array<T>& x);
template<class T> void rndInt(MT::Array<T>& a,int low=0,int high=1);
template<class T> void rndUni(MT::Array<T>& a,double low=0.,double high=1.,bool add=false);
template<class T> void rndGauss(MT::Array<T>& a,double stdDev=1.,bool add=false);
//template<class T> void rndGauss(MT::Array<T>& a,bool add=false);
//template<class T> MT::Array<T>& rndGauss(double stdDev,uint dim);
template<class T> uint softMax(const MT::Array<T>& a,arr& soft,double beta);
template<class T> void grid(MT::Array<T>& a,uint dim,T lo,T hi,uint steps);

template<class T> void writeArr(const MT::Array<T>& x,const char* tag,std::ofstream& os);
template<class T> bool readArr(MT::Array<T>& x,const char *tag,std::ifstream& is);


//===========================================================================

inline uintA TUP(uint i){                                           uintA z(1); z(0)=i; return z; }
inline uintA TUP(uint i,uint j){                                    uintA z(2); z(0)=i; z(1)=j; return z; }
inline uintA TUP(uint i,uint j,uint k){                             uintA z(3); z(0)=i; z(1)=j; z(2)=k; return z; }
inline uintA TUP(uint i,uint j,uint k,uint l){                      uintA z(4); z(0)=i; z(1)=j; z(2)=k; z(3)=l; return z; }
inline uintA TUP(uint i,uint j,uint k,uint l,uint m){               uintA z(5); z(0)=i; z(1)=j; z(2)=k; z(3)=l; z(4)=m; return z; }
inline uintA TUP(uint i,uint j,uint k,uint l,uint m,uint n){        uintA z(6); z(0)=i; z(1)=j; z(2)=k; z(3)=l; z(4)=m; z(5)=n; return z; }
inline uintA TUP(uint i,uint j,uint k,uint l,uint m,uint n,uint o){ uintA z(7); z(0)=i; z(1)=j; z(2)=k; z(3)=l; z(4)=m; z(5)=n; z(6)=o; return z; }
inline uintA TUP(uint i,uint j,uint k,uint l,uint m,uint n,uint o,uint p){ uintA z(8); z(0)=i; z(1)=j; z(2)=k; z(3)=l; z(4)=m; z(5)=n; z(6)=o; z(7)=p; return z; }

inline arr ARR(double i){                                                  arr z(1); z(0)=i; return z; }
inline arr ARR(double i,double j){                                        arr z(2); z(0)=i; z(1)=j; return z; }
inline arr ARR(double i,double j,double k){                             arr z(3); z(0)=i; z(1)=j; z(2)=k; return z; }
inline arr ARR(double i,double j,double k,double l){                      arr z(4); z(0)=i; z(1)=j; z(2)=k; z(3)=l; return z; }
inline arr ARR(double i,double j,double k,double l,double m){               arr z(5); z(0)=i; z(1)=j; z(2)=k; z(3)=l; z(4)=m; return z; }
inline arr ARR(double i,double j,double k,double l,double m,double n){        arr z(6); z(0)=i; z(1)=j; z(2)=k; z(3)=l; z(4)=m; z(5)=n; return z; }
inline arr ARR(double i,double j,double k,double l,double m,double n,double o){ arr z(7); z(0)=i; z(1)=j; z(2)=k; z(3)=l; z(4)=m; z(5)=n; z(6)=o; return z; }


//===========================================================================
//
//!@name basic operators
//

namespace MT{
  template<class T> Array<T> operator~(const Array<T>& y);
  template<class T> Array<T> operator-(const Array<T>& y);
  template<class T> Array<T> operator^(const Array<T>& y,const Array<T>& z);
  template<class T> Array<T> operator*(const Array<T>& y,const Array<T>& z);
  template<class T> Array<T> operator*(const Array<T>& y,T z);
  template<class T> Array<T> operator*(T y,const Array<T>& z);

#define BinaryOperator( op, name)					\
  template<class T> Array<T> operator op(const Array<T>& y,const Array<T>& z); \
    template<class T> Array<T> operator op(T y,const Array<T>& z);	\
    template<class T> Array<T> operator op(const Array<T>& y,T z);
  BinaryOperator( + , plusA)
    BinaryOperator( - , minusA)
    BinaryOperator( % , mult)
    BinaryOperator( / , div)
#undef BinaryOperator

#define CompoundAssignmentOperator( op )				\
    template<class T> Array<T>& operator op (Array<T>& x,const Array<T>& y); \
      template<class T> Array<T>& operator op ( Array<T>& x, T y );
    CompoundAssignmentOperator( |= )
      CompoundAssignmentOperator( ^= )
      CompoundAssignmentOperator( &= )
      CompoundAssignmentOperator( += )
      CompoundAssignmentOperator( -= )
      CompoundAssignmentOperator( *= )
      CompoundAssignmentOperator( /= )
      CompoundAssignmentOperator( %= )
#undef CompoundAssignmentOperator
      }

#define UnaryOperation( name, op )					\
  template<class T> MT::Array<T>& name (MT::Array<T>& x,const MT::Array<T>& y); 
UnaryOperation( negative, - )
#undef UnaryOperation

#define BinaryOperation( name, op )					\
  template<class T> MT::Array<T>& name(MT::Array<T>& x,const MT::Array<T>& y,const MT::Array<T>& z); \
  template<class T> MT::Array<T>& name##S(MT::Array<T>& x,const MT::Array<T>& y,T z); \
  template<class T> MT::Array<T>& name##S(MT::Array<T>& x, T y, const MT::Array<T>& z);
  BinaryOperation( plusA , + )
    BinaryOperation( minusA , - )
    BinaryOperation( mult , * )
    BinaryOperation( div , / )
#undef BinaryOperation

#define UnaryFunction( func )						\
    template<class T> MT::Array<T> func (const MT::Array<T>& y);
    UnaryFunction( acos )
      UnaryFunction( asin )
      UnaryFunction( atan )
      UnaryFunction( cos )
      UnaryFunction( sin )
      UnaryFunction( tan )
      UnaryFunction( cosh )
      UnaryFunction( sinh )
      UnaryFunction( tanh )
      UnaryFunction( acosh )
      UnaryFunction( asinh )
      UnaryFunction( atanh )
      UnaryFunction( exp )
      UnaryFunction( log )
      UnaryFunction( log10 )
      UnaryFunction( sqrt )
      UnaryFunction( cbrt )
      UnaryFunction( ceil )
      UnaryFunction( fabs )
      UnaryFunction( floor )
#undef UnaryFunction

#define BinaryFunction( func )						\
      template<class T> MT::Array<T> func(const MT::Array<T>& y,const MT::Array<T>& z); \
      template<class T> MT::Array<T> func(const MT::Array<T>& y,T z);	\
      template<class T> MT::Array<T> func(T y,const MT::Array<T>& z);
      BinaryFunction( atan2 )
	BinaryFunction( pow )
	BinaryFunction( fmod )
#undef BinaryFunction

	template<class T> std::istream& operator>>(std::istream& is,MT::Array<T>& x);
template<class T> MT::Array<T>& operator<<(MT::Array<T>& x,const char* str);
template<class T> std::ostream& operator<<(std::ostream& os,const MT::Array<T>& x);
template<class T,class S> bool samedim(const MT::Array<T>& a,const MT::Array<S>& b);
template<class T> void checkNan(const MT::Array<T>& x);
template<class T> bool operator==(const MT::Array<T>& v,const MT::Array<T>& w);
template<class T> bool operator==(const MT::Array<T>& v,const T *w);
template<class T> bool operator!=(const MT::Array<T>& v,const MT::Array<T>& w);
template<class T> bool operator<(const MT::Array<T>& v,const MT::Array<T>& w);





  
//===========================================================================
//
// implementations
//

#if 1 //MT_IMPLICIT_TEMPLATES
#  include"array_t.cpp"
#endif
#ifdef MT_IMPLEMENTATION
#  include"array.cpp"
#endif

#endif
