/* Copyright (C) 2000, 2006  Marc Toussaint (mtoussai@inf.ed.ac.uk)
    under the terms of the GNU LGPL (http://www.gnu.org/copyleft/lesser.html)
    see the `std.h' file for a full copyright statement  */

#ifndef MT_inference_h
#define MT_inference_h

#include "array.h"

#define INFER_DEBUG_LEVEL 0

#define INFER_DEBUG(l,x) if(l<=INFER_DEBUG_LEVEL) x;

extern bool MaxMarginalMode;

typedef void (*linearizationT)(const arr& x,arr& f,arr& F,arr& Q);

//===========================================================================

//! possible types of variables
enum VType { dis, cont };
struct Variable{
  uint d;
  MT::String name;
  void operator=(const Variable& v){
    d=v.d;
    name=v.name;
  }
};
inline ostream& operator<<(ostream& os,const Variable& v){
  os <<v.name <<'(' <<v.d <<')';
  return os;
}
typedef MT::Array<Variable> variableA;


//! possible types of factors \ingroup infer2
enum FType { TableFT, SwitchFT, GaussFT, MogFT, UTransFT, LinearFT, GaussFctFT };
extern const char *FTypeNames[6];

/*! generic probabilistic factor -- mainly virtuals \ingroup infer2 */
class Factor{
public:
  FType type; //!< specifies the type of this factor
  uintA id;   //!< f=f(x_1,x_3,x_7) => id=[1,3,7]; array of variables (their indices) this factor depends on
  uintA dim;  //!< f=f(x_1,x_3,x_7) => dim=[dim(x_1),dim(x_3),dim(x_7)];  array of dimensionalities of the variables this factor depends on

  //Factor(uintA &ids,uintA &dim){ NIY; }
  virtual ~Factor(){}

  void setBase(FType _type,const uintA &_var,const uintA &_dim){ type=_type; id=_var; dim=_dim; }

  virtual Factor *clone() = 0;                   //!< allocates factor of same type -- remember to delete!!
  //virtual Factor *new_maginal(uintA& mid) = 0;   //!< allocates a factor with the proper type for the marginal (mid)
  virtual bool operator==(Factor& q) = 0;        //!< check if f==q
  virtual void random(){ NIY };                  //!< randomize f (for automatic (unit) testing)

  virtual void uniform() = 0;                    //!< f(x) = 1
  virtual void set(const Factor& q) = 0;         //!< f(x) = q(x)
  virtual void set(const arr& q){ NIY; };        //!< f(x) = q(x)
  virtual void multiply(const Factor& q) = 0;    //!< f(x) = f(x) * q(x)
  virtual void invMultiply(const Factor& q){ NIY; }; //!< f(x) = q(x) / f(x)
  virtual void divide(const Factor& q){ NIY; };  //!< f(x) = f(x) / q(x)
  //virtual void multiply(const Factor& m,uintA& mid) = 0; //!< f(y,z) = f(y,z) * q(y);  \c i determines indices of `y'
  virtual void marginal(Factor*& m,uintA& mid) = 0; //!< m(y) = sum_z f(y,z);  \c i determines indices of `y', e.g., f=f(x_1,x_3,x_7), i=[3,7] => m(x_3,x_7) = sum_{x_1} f(x_1,x_3,x_7)
  virtual void marginal(Factor & m){ NIY; };     //!< m(y) = sum_z f(y,z);  \c i determines indices of `y', e.g., f=f(x_1,x_3,x_7), i=[3,7] => m(x_3,x_7) = sum_{x_1} f(x_1,x_3,x_7)
  virtual void evidence(uint e){ NIY; };         //!< f(e) = 1,  f(x!=e) = 0
  virtual void evidence(const arr& e){ NIY; };
  virtual void sample(uint *s=0,arr *x=0){ NIY };//!< sample (s=discrete, x=continuous)

  virtual void write(std::ostream& os) const = 0;//!< writes factor to the ostream
};
stdOutPipe(Factor);

Factor *new_factor(FType type,variableA& V,uintA& ids);


//===========================================================================
//
// basic data structures: FactorGraph
//

typedef Factor *FactorP;
class TableFac;

//! struct stored at each node of the factor graph \ingroup infer2
struct factorNode{
  Factor *f,*org;   //!< storing the factor and perhaps the initialization factor (to check consistency)
  MT::String name;  //!< a name for easier access
  factorNode& data(){ return *this; }
  factorNode(){ f=org=0; }
  ~factorNode(){ freemem(); }
  void freemem(){ if(f) delete f; if(org) delete org; f=org=0; }
  Factor *operator->() const{ return f; }
  operator Factor*(){ return f; }
  void operator=(Factor *_f){ f=_f; }
  void operator=(const Factor &_f){ *f=_f; }
  void operator=(const factorNode& n){
    freemem();
    if(n.f) f=n.f->clone(); if(n.org) org=n.org->clone();
    name=n.name;
  }
};

//! struct stored at each edge of the factor graph \ingroup infer2
struct factorEdge{
  Factor *fwd,*bwd; //!< storing the accumulation of all passed forward and backward messages
  bool temporal;    //!< flag if this is a temporal edge (used in unrollDBN)
  int label;        //!< a label for easier access
  uint fr,to;
  factorEdge& data(){ return *this; }
  factorEdge(){ fwd=bwd=0; temporal=false; label=0; }
  ~factorEdge(){ freemem(); }
  void freemem(){ if(fwd) delete fwd; if(bwd) delete bwd; fwd=bwd=0; }
  void operator=(const factorEdge& e){
    freemem();
    if(e.fwd) fwd=e.fwd->clone(); if(e.bwd) bwd=e.bwd->clone();
    temporal=e.temporal;
    label=e.label;
    fr=e.fr;
    to=e.to;
  }
  void operator=(Factor *_f){ fwd=_f; }
};

inline ostream& operator<<(ostream& os,const factorNode& fn){
  os <<FTypeNames[fn->type] <<'(' <<fn->id <<')' <<endl;
  fn->write(os);
  return os;
}
inline ostream& operator<<(ostream& os,const factorEdge& fe){
  if(fe.fwd){
    os <<FTypeNames[fe.fwd->type] <<'(' <<fe.fwd->id <<')' <<endl;
    fe.fwd->write(os);
  }
  if(fe.bwd){
    os <<endl;
    os <<FTypeNames[fe.bwd->type] <<'(' <<fe.bwd->id <<')' <<endl;
    fe.bwd->write(os);
  }
  return os;
}

class FactorModel:public MT::Array<factorNode>{
public:
  MT::Array<factorEdge> E;
  MT::Array<Variable> V;

  uint T;//whether int contains multiple time slices (DBN)
  uint dF,dV; //factors/variables per time-slice

  FactorModel& operator=(const FactorModel& M){
    resize(M.N);
    for(uint i=0;i<N;i++) elem(i) = M.elem(i);
    E=M.E;
    V=M.V;
    return *this;
  }

  void clear(){ E.clear(); V.clear(); ((MT::Array<factorNode>*)this)->clear(); }
  void clearFactors(){ E.clear(); ((MT::Array<factorNode>*)this)->clear(); }

  uint variable(uint d,const char* name){
    V.resizeCopy(V.N+1);
    V(V.N-1).name=name;
    V(V.N-1).d=d;
    return V.N-1;
  }

  uint factor(FType type,uintA var){
    resizeCopy(N+1);
    factorNode *n=&elem(N-1);
    uintA dim(var.N);
    for(uint i=0;i<var.N;i++) dim(i)=V(var(i)).d;
    n->f=new_factor(type,V,var);
    return N-1;
  }

  template<class T>
  uint factor(FType type,uintA ids,const T &init){
    factor(type,ids);
    elem(N-1)->set(init);
    return N-1;
  }

  factorEdge &arc(uint i,uint j,bool temporal=false){
    E.resizeCopy(E.N+1);
    E(E.N-1).fr=i;
    E(E.N-1).to=j;
    E(E.N-1).temporal=temporal;
    return E(E.N-1);
  }

  void arcs(const uintA &list){
    CHECK(!(list.N%2),"");
    uint i;
    E.resize(list.N/2);
    for(i=0;i<list.N;i+=2){
      E(i/2).fr=list.elem(i);
      E(i/2).to=list.elem(i+1);
    }
  }

  Factor *from(uint i){ return elem(E(i).fr); }
  Factor *to  (uint i){ return elem(E(i).to); }

  void write(std::ostream& os) const{
    uint i;
    os <<"<FactorModel>"
       <<" #variable=" <<V.N
       <<" #factors=" <<N
       <<" #arcs=" <<E.N <<endl;
    for(i=0;i<V.N;i++){
      os <<'[' <<i <<"] " <<V.elem(i) <<endl;
    }
    os <<endl;
    for(i=0;i<N;i++){
      os <<'[' <<i <<"] ";
      os <<FTypeNames[elem(i)->type] <<'(' <<elem(i)->id <<')' <<endl;
      os <<*elem(i) <<endl;
    }
    os <<endl;
    for(i=0;i<E.N;i++) os <<'[' <<E(i).fr <<'-' <<E(i).to <<"] " <<E(i) <<endl;
    os <<"\n</FactorModel>" <<endl;
  }
};
stdOutPipe(FactorModel);

void allocateEdgeFactors(FactorModel& M);
void storeOrgFactors(FactorModel& M);
void checkConsistency(FactorModel& M);
bool checkEquilibrium(FactorModel& M);
void checkConsistency(FactorModel& M,uint i);
void passLabeledEdges(FactorModel& M,int label,FwdBwd dir);
void unrollDBN(FactorModel& M,uint T,uint vars_per_slice);
void getIndices(uintA& list,const uintA& id,const uintA& dim,const uintA& mid);
void getPick(uintA& pick,const uintA& id,const uintA& mid);


//===========================================================================
//
// special Factor Types
//

//*! A factor over discrete variables represented as table. \ingroup infer2 */
class TableFac:public Factor{
public:
  arr P;        //!< the (probability) table
  double logP;  //!< the log-scaling of the table, such that true_factor = exp(logP) * P

  TableFac();
  //TableFac(uint _var,uint _dim);
  //TableFac(uint _var,uint _dim,arr& q);
  TableFac(const uintA &_var,const uintA &_dim);
  TableFac(const uintA &_var,const uintA &_dim,const arr &q);
  TableFac(const variableA &V,const uintA &_var);
  TableFac(const variableA &V,const uintA &_var,const arr &q);
  /*
  TableFac(uint d0,uint i0);
  TableFac(uint d0,uint i0,const arr& q);
  TableFac(uint d0,uint i0,uint d1,uint i1);
  TableFac(uint d0,uint i0,uint d1,uint i1,const arr& q);
  TableFac(uint d0,uint i0,uint d1,uint i1,uint d2,uint i2,const arr& q);
  */

  virtual void random();

  virtual void uniform();
  virtual void multiply(const Factor& q);
  virtual void invMultiply(const Factor& q);
  virtual void set(const Factor& q);
  virtual void set(const arr& q);
  virtual void divide(const Factor& q);
  //virtual void multiply(const Factor& m,uintA& mid);
  virtual void evidence(uint e);
  virtual void evidence(const arr& e);
  virtual void marginal(Factor*& m,uintA& mid);
  virtual void marginal(Factor& m);
  virtual void write(std::ostream& os) const;
  virtual Factor* clone();
  virtual bool operator==(Factor& q);

  void weightedAdd(double w,const Factor& q);
};

//*! A deterministic switch factor over a number of discrete variables. \ingroup infer2 */
class SwitchFac:public Factor{
public:
  uint d,n;
  arr Pi;
  arr P;        //!< the (probability) table
  double Z;     //!< the log-scaling of the table, such that true_factor = exp(logP) * P

  SwitchFac();
  SwitchFac(const variableA &V,const uintA &_var);
  SwitchFac(const variableA &V,const uintA &_var,const arr &q);

  virtual Factor* clone();
  virtual void set(const Factor& q);
  virtual void set(const arr& q);
  virtual bool operator==(Factor& q);
  virtual void random();
  virtual void uniform();

  virtual void multiply(const Factor& q);
  virtual void marginal(Factor& m);
  virtual void marginal(Factor*& m,uintA& mid);

  virtual void write(std::ostream& os) const;
};



#ifdef MT_IMPLEMENTATION
#  include"infer_old.cpp"
#endif

#endif
